#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <netinet/in.h>
#include <string.h>

#include "pair.h"

void print_help() {
	 printf("Network Host Pairing v1.0\n\n");
	 printf("Run this tool on a set of machines and see them paired up.\n");
	 printf("It uses multicast to find potential \"partners\".\n\n");
	 printf(" -m 224.0.0.99 : obtain host list via said multicast group\n");
	 printf(" -a            : always play 'active' (default: random)\n");
	 printf(" -s            : always play 'passive' (default: random)\n");
	 printf(" -r <seconds>  : timeout between re-evaluating role (when random), default 20\n");
	 printf(" -i <seconds>  : timeout waiting for ICMP replies, default 5\n");
	 printf(" -t <seconds>  : limit on total time spent, default 120\n");
	 printf(" -p <port>     : TCP port to use for negotiations, default 8899\n");
	 printf(" -h            : print help\n");
	 printf(" -v            : print version\n");
	return;
}


/*
 * 1. decide on active/passive role
 * 2. get partner list, if active; otherwise, start listening
 * 3. start appropriate task
 * 4. if timeout reached, goto 1
 */
int main(int argc, char **argv) {
	int role_passive=0; /* 0 for active, 1 for passive */
	int role_random=1;
	struct timeval curtime;
	int opt;
	int timeout_total=120; /* global t/o in seconds */
	int timeout_role=20; /* role changing t/o in seconds */
	int timeout_icmp=5; /* timeout waiting for ICMP replies to arrive */
	struct in_addr hostlist[MAX_HOSTS]; /* potential partners' list */
	int numhosts=0; /* elements in hostlist */
	char maddr[16]; /* multicast address as a string */
	int port=8899; /* tcp port no. for negotiations */
	time_t starttime;
	struct in_addr peer; /* our partner */

	extern char *optarg;
	extern int optopt;

	memset(maddr,0,16);
	snprintf(maddr,16,"224.0.0.99");

	while((opt = getopt(argc, argv, "hvasp:m:t:r:i:")) != -1) {
		switch(opt) {
			case 'h':
				print_help();
				exit(0);
				break;
			case 'v':
				printf("v1.2.3.4 FIXME\n");
				exit(0);
				break;
			case 'a':
				if (role_random==1) {
					role_passive=0;
					role_random=0;
				} else {
					printf("Error: -a and -s are mutually exclusive.\n");
					print_help();
					exit(1);
				}
				break;
			case 's':
				if (role_random==1) {
					role_passive=1;
					role_random=0;
				} else {
					printf("Error: -a and -s are mutually exclusive.\n");
					print_help();
					exit(1);
				}
				break;
			case 'p':
				port=atoi(optarg);
				if (port<=0) {
					printf("Error: -p: positive integer value required.\n");
					print_help();
					exit(1);
				}
				break;
			case 'm':
				snprintf(maddr,16,"%s",optarg);
				break;
			case 't':
				timeout_total=atoi(optarg);
				if(timeout_total<=0) {
					printf("Error: -t: positive integer value required, %d.\n",timeout_total);
					print_help();
					exit(1);
				}
				break;
			case 'r':
				timeout_role=atoi(optarg);
				if(timeout_role<=0) {
					printf("Error: -r: positive integer value required.\n");
					print_help();
					exit(1);
				}
				break;
			case 'i':
				timeout_icmp=atoi(optarg);
				if(timeout_icmp<=0) {
					printf("Error: -i: positive integer value required.\n");
					print_help();
					exit(1);
				}
				break;

			case ':':
				printf("Error: option `%c' needs a value\n\n", optopt);
				print_help();
				exit(1);
				break;
			case '?':
				printf("Error: no such option: `%c'\n\n", optopt);
				print_help();
				exit(1);
			}
		}

	starttime=time(NULL);

	do {
		/* poor man's random number: we'd need to seed rand() anyway and
		 * it's not important go get "real" randomness in here */
		if (role_random==1) {
			gettimeofday( &curtime, NULL );
			role_passive=curtime.tv_usec % 2;
		}

		/* start pairing */
		if (role_passive) {
			printf("-- passive mode enabled --\n");
			/* do a 'passive mode' stunt for timeout_role seconds */
			if(0<passive(port,timeout_role,maddr,&peer)) {
				printf("ACTIVE PEER is %s\n",inet_ntoa(peer));
				exit(0);
			}
		} else {
			/* get the list of potential 'partners' */
			printf("-- active mode enabled --\n");
			build_host_list(hostlist, &numhosts, maddr, timeout_icmp);
			/* do an 'active mode' stunt for timeout_role seconds */
			if(0<active(hostlist,numhosts,port,timeout_role,&peer)) {
				printf("PASSIVE PEER is %s\n",inet_ntoa(peer));
				exit(0);
			}
		}
	} while(starttime+timeout_total>time(NULL));

	printf("NO PARTNERS FOUND\n");
	exit(1);

	return 0;
}
