#include <fcntl.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>

#include "pair.h"

/* passive role: listen on given port on all interfaces */
int passive(int port, int timeout, const char *maddr, struct in_addr *peer) {

	int sockd, sockd2; /* the tcp socket, and accept'ed socket */
	int msockd; /* the udp socket for multicast pings, see below */
	socklen_t addrlen;
	struct sockaddr_in my_name, peer_name;
	int status;
	struct timeval listen_timeout;
	fd_set rset;
	struct linger mylinger;

	const char *proto = "HEY"; /* this is sent to the connection "winner" */

	/* on newer kernels this is 0 by default -> change that while we run */
	long old_icmp_status;
	const char *icmp_sysctl_path="/proc/sys/net/ipv4/icmp_echo_ignore_broadcasts";

	unsigned int loop=0; 
	struct ip_mreq mc; /* multicast stuff */

	/* create tcp socket */
	sockd = socket(AF_INET, SOCK_STREAM, 0);
	if (sockd == -1)
	{
		perror("passive: tcp socket creation error");
		exit(1);
	}

	status=fcntl(sockd,F_SETFL,O_NONBLOCK);
	if (status == -1)
	{
		perror("passive: fcntl()");
		exit(1);
	}

	/* drop all data on close, avoid TIME_WAIT */
	mylinger.l_onoff=1;
	mylinger.l_linger=0;
	setsockopt(sockd,SOL_SOCKET,SO_LINGER,&mylinger,sizeof(struct linger));

	/* create udp socket too -- apparently one can't even send/receive
	 * ICMP/multicast on a socket opened for SOCK_STREAM 
	 */
	msockd = socket(AF_INET, SOCK_DGRAM, 0);
	if (msockd == -1)
	{
		perror("passive: udp socket creation error");
		exit(1);
	}

	/* no multicast loopback, thank you */
	status = setsockopt(msockd,IPPROTO_IP,IP_MULTICAST_LOOP,&loop,sizeof(int));
	if (status == -1)
	{
		perror("passive: IP_MULTICAST_LOOP");
		exit(1);
	}

	/* Join multicast group.
	 * This is enough as we are waiting for ICMP ECHO here,
	 * the kernel will reply on its own.
	 * */
	printf("joining multicast group %s\n",maddr);
	mc.imr_multiaddr.s_addr=inet_addr(maddr);
	mc.imr_interface.s_addr=INADDR_ANY; /* any interface */
	status = setsockopt(msockd,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mc,sizeof(struct ip_mreq));
	if (status == -1)
	{
		perror("passive: IP_ADD_MEMBERSHIP");
		exit(1);
	}

	/* server address */
	my_name.sin_family = AF_INET;
	my_name.sin_addr.s_addr = INADDR_ANY;
	my_name.sin_port = htons(port);

	status = bind(sockd, (struct sockaddr*)&my_name, sizeof(struct sockaddr_in));
	if (status == -1)
	{
		perror("passive: binding error");
		exit(1);
	}

	/* 0 backlog - only 1 tcp connection allowed (the winner 'active' peer) */
	status = listen(sockd, 0);
	if (status == -1)
	{
		perror("passive: listening error");
		exit(1);
	}

	/* enable replies to ICMP on multicast */
	old_icmp_status=get_sysctl_status(icmp_sysctl_path);
	if(0!=old_icmp_status) {
		printf("icmp_echo_ignore_broadcasts == %ld, setting it to 0\n", old_icmp_status);
		set_sysctl_status(icmp_sysctl_path, 0);
	}

	/* wait for a tcp connection */
	printf("listening on %d, timeout %d\n",port,timeout);
	addrlen = sizeof(struct sockaddr_in);
	FD_ZERO(&rset);
	FD_SET(sockd,&rset);
	listen_timeout.tv_sec=timeout;
	listen_timeout.tv_usec=0;

	status=select(sockd+1,&rset,NULL,NULL,&listen_timeout);
	if(status == -1) {
		perror("passive: select()");
	} else {
			/* was this a 'real' connection ?*/
			if (FD_ISSET(sockd, &rset)) {
				sockd2 = accept(sockd, (struct sockaddr*)&peer_name, &addrlen);
				if (sockd2 == -1)
				{
					perror("passive: wrong connection");
				}
				/* we have a valid &peer_name ! */

				/* confirm selection */
				write(sockd2,proto,3);

				close(sockd2);
				close(sockd);
			} else {
				/* false alarm, exiting anyway, with error */
				printf("leaving multicast group %s\n",maddr);
				status = setsockopt(msockd,IPPROTO_IP,IP_DROP_MEMBERSHIP,&mc,sizeof(struct ip_mreq));
				if (status == -1)
				{
					perror("passive: IP_DROP_MEMBERSHIP");
					exit(1);
				}
				close(sockd);
				close(msockd);
				return(-1);
			}
	}

	printf("leaving multicast group %s\n",maddr);
	status = setsockopt(msockd,IPPROTO_IP,IP_DROP_MEMBERSHIP,&mc,sizeof(struct ip_mreq));
	if (status == -1)
	{
		perror("passive: IP_DROP_MEMBERSHIP");
		exit(1);
	}
	close(msockd);

	/* clean up ICMP reply status */
	if(0!=old_icmp_status) {
		printf("restoring icmp_echo_ignore_broadcasts value\n");
		set_sysctl_status(icmp_sysctl_path, old_icmp_status);
	}
		
	/* ip address of peer */
	peer->s_addr=peer_name.sin_addr.s_addr;
	return(1);
}
